from datetime import timezone
from typing import Union
from zoneinfo import ZoneInfo

import pytest
from parameterized import parameterized  # type: ignore[import]
from pydantic import BaseModel, ValidationError

from heliclockter import DateTimeTzT, datetime_local, datetime_tz, datetime_utc, timedelta

from tests.shared import datetime_cet


class DatetimeTZModel(BaseModel):
    dt: datetime_tz


class DatetimeUTCModel(BaseModel):
    dt: datetime_utc


class DatetimeCETModel(BaseModel):
    dt: datetime_cet


class DatetimeLocalModel(BaseModel):
    dt: datetime_local


class DatetimeDefaultObject(BaseModel):
    dt_now: datetime_tz = datetime_utc.now()
    dt: str = datetime_utc.future(days=120).isoformat()


TestModelT = Union[DatetimeTZModel, DatetimeUTCModel, DatetimeCETModel]


@parameterized.expand(
    [
        # UTC tests
        (
            '2021-01-10T10:00:00',
            datetime_utc(2021, 1, 10, 10, 00, 00, tzinfo=ZoneInfo('UTC')),
            DatetimeUTCModel,
        ),
        (
            '2021-01-10T10:00:00+04:00',
            datetime_utc(2021, 1, 10, 6, 00, 00, tzinfo=ZoneInfo('UTC')),
            DatetimeUTCModel,
        ),
        (
            '2021-01-10T10:00:00+00:00',
            datetime_utc(2021, 1, 10, 10, 00, 00, tzinfo=ZoneInfo('UTC')),
            DatetimeUTCModel,
        ),
        (
            '2021-01-10T10:00:00-04:00',
            datetime_utc(2021, 1, 10, 14, 00, 00, tzinfo=ZoneInfo('UTC')),
            DatetimeUTCModel,
        ),
        # TZ tests
        (
            '2021-01-10T10:00:00+04:00',
            datetime_tz(2021, 1, 10, 10, 0, tzinfo=timezone(timedelta(hours=4))),
            DatetimeTZModel,
        ),
        (
            '2021-01-10T10:00:00-04:00',
            datetime_tz(2021, 1, 10, 10, 0, tzinfo=timezone(timedelta(hours=-4))),
            DatetimeTZModel,
        ),
        # CET tests
        (
            '2021-01-10T10:00:00',
            datetime_cet(2021, 1, 10, 10, 00, 00, tzinfo=ZoneInfo('CET')),
            DatetimeCETModel,
        ),
        (
            '2021-01-10T10:00:00+04:00',
            datetime_cet(2021, 1, 10, 7, 00, 00, tzinfo=ZoneInfo('CET')),
            DatetimeCETModel,
        ),
        (
            '2021-01-10T10:00:00+00:00',
            datetime_cet(2021, 1, 10, 11, 00, 00, tzinfo=ZoneInfo('CET')),
            DatetimeCETModel,
        ),
        (
            '2021-01-10T10:00:00-04:00',
            datetime_cet(2021, 1, 10, 15, 00, 00, tzinfo=ZoneInfo('CET')),
            DatetimeCETModel,
        ),
    ]
)
def test_datetime_parsing(test_str: str, expectation: DateTimeTzT, model: TestModelT) -> None:
    parsed_model = model.parse_obj({'dt': test_str})
    assert isinstance(parsed_model.dt, type(expectation))
    assert parsed_model.dt == expectation


def test_datetime_local_parsing() -> None:
    parsed_model = DatetimeLocalModel.parse_obj({'dt': '2021-01-10T10:00:00-04:00'})
    assert isinstance(parsed_model.dt, datetime_local)


def test_create_default_pydantic_field() -> None:
    obj = DatetimeDefaultObject()
    assert obj.dt
    assert obj.dt_now


def test_parse_datetime_utc_as_datetime_tz() -> None:
    obj = DatetimeDefaultObject(dt_now=datetime_utc.now())
    assert isinstance(obj.dt_now, datetime_tz)


def test_parse_datetime_tz_without_timezone() -> None:
    with pytest.raises(ValidationError):
        DatetimeTZModel.parse_obj({'dt': '2021-01-10T10:00:00'})



"""
We cannot write unittest using `datetime_local` that asserts a fixed offset compared to some
other datetime_tz class because the local timezone depends on the machine that runs the tests.

Because of this we declare `datetime_cet` here to test with.
"""
from zoneinfo import ZoneInfo

from heliclockter import datetime_tz


class datetime_cet(datetime_tz):
    """
    A `datetime_cet` is a `datetime_tz` but which is guaranteed to be in the 'CET' timezone.
    """

    assumed_timezone_for_timezone_naive_input = ZoneInfo('CET')



from __future__ import annotations

import datetime as _datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Iterator,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)
from zoneinfo import ZoneInfo

# We don't require pydantic as a dependency, but add validate logic if it exists.
try:
    from pydantic.datetime_parse import parse_datetime

    PYDANTIC_AVAILABLE = True
except ImportError:
    PYDANTIC_AVAILABLE = False


# `date` and `timedelta` are exposed for your convenience in case this module is used in combination
# with an import linter that prohibits importing the `datetime` package anywhere.
date = _datetime.date
timedelta = _datetime.timedelta


tz_local = cast(ZoneInfo, _datetime.datetime.now().astimezone().tzinfo)

__version__ = '1.0.0'


DateTimeTzT = TypeVar('DateTimeTzT', bound='datetime_tz')
IntFloat = Union[int, float]


class DatetimeTzError(ValueError):
    """
    An error with the input value when trying to create or mutate a `datetime_tz` instance.
    """


class datetime_tz(_datetime.datetime):
    """
    A `datetime_tz` is just a `datetime.datetime` but which is guaranteed to be timezone aware.
    """

    assumed_timezone_for_timezone_naive_input: ClassVar[Optional[ZoneInfo]] = None

    if TYPE_CHECKING:

        def __init__(
            self,
            year: int,
            month: int,
            day: int,
            hour: int = 0,
            minute: int = 0,
            second: int = 0,
            microsecond: int = 0,
            *,
            tzinfo: _datetime.tzinfo,
        ) -> None:
            pass

    else:

        def __init__(  # pylint: disable=unused-argument
            self,
            year: int,
            month: int,
            day: int,
            hour: int = 0,
            minute: int = 0,
            second: int = 0,
            microsecond: int = 0,
            tzinfo: _datetime.tzinfo = None,
        ) -> None:
            msg = f'{self.__class__} must have a timezone'
            assert tzinfo is not None and self.tzinfo is not None, msg
            tz_expected = self.assumed_timezone_for_timezone_naive_input or tzinfo

            msg = f'{self.__class__} got invalid timezone {self.tzinfo!r}, expected {tz_expected!r}'
            assert self.tzinfo == tz_expected, msg

            self.assert_aware_datetime(self)

    if PYDANTIC_AVAILABLE:

        @classmethod
        def __get_validators__(cls) -> Iterator[Callable[[Any], Optional[datetime_tz]]]:
            yield cls._validate

        @classmethod
        def _validate(cls: Type[DateTimeTzT], v: Any) -> Optional[DateTimeTzT]:
            return cls.from_datetime(parse_datetime(v)) if v else None

    @classmethod
    def from_datetime(cls: Type[DateTimeTzT], dt: _datetime.datetime) -> DateTimeTzT:
        # Case datetime is naive and there is no assumed timezone.
        if dt.tzinfo is None and cls.assumed_timezone_for_timezone_naive_input is None:
            raise DatetimeTzError('Cannot create aware datetime from naive if no tz is assumed')

        # Case: datetime is naive, but the timezone is assumed.
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=cls.assumed_timezone_for_timezone_naive_input)

        # Case: datetime is aware and the timezone is assumed, enforce that timezone.
        elif (assumed_tz := cls.assumed_timezone_for_timezone_naive_input) is not None:
            dt = dt.astimezone(assumed_tz)

        cls.assert_aware_datetime(dt)
        return cls(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            microsecond=dt.microsecond,
            tzinfo=dt.tzinfo,  # type: ignore[arg-type]
        )

    @classmethod
    def now(cls: Type[DateTimeTzT], tz: Optional[_datetime.tzinfo] = None) -> DateTimeTzT:
        tz = cls.assumed_timezone_for_timezone_naive_input or tz
        if tz is None:
            raise DatetimeTzError(
                'Must override assumed_timezone_for_timezone_naive_input '
                'or give a timezone when calling now'
            )
        return cls.from_datetime(_datetime.datetime.now(tz))

    @classmethod
    def future(
        cls: Type[DateTimeTzT],
        weeks: IntFloat = 0,
        days: IntFloat = 0,
        hours: IntFloat = 0,
        minutes: IntFloat = 0,
        seconds: IntFloat = 0,
        milliseconds: IntFloat = 0,
        microseconds: IntFloat = 0,
        tz: Optional[ZoneInfo] = None,
    ) -> DateTimeTzT:
        delta = timedelta(
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            milliseconds=milliseconds,
            microseconds=microseconds,
        )
        return cls.now(tz=tz) + delta

    @classmethod
    def past(
        cls: Type[DateTimeTzT],
        weeks: IntFloat = 0,
        days: IntFloat = 0,
        hours: IntFloat = 0,
        minutes: IntFloat = 0,
        seconds: IntFloat = 0,
        milliseconds: IntFloat = 0,
        microseconds: IntFloat = 0,
        tz: Optional[ZoneInfo] = None,
    ) -> DateTimeTzT:
        delta = timedelta(
            weeks=weeks,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            milliseconds=milliseconds,
            microseconds=microseconds,
        )
        return cls.now(tz=tz) - delta

    @classmethod
    def fromisoformat(cls: Type[DateTimeTzT], date_string: str) -> DateTimeTzT:
        return cls.from_datetime(_datetime.datetime.fromisoformat(date_string))

    @classmethod
    def strptime(cls: Type[DateTimeTzT], date_string: str, __format: str) -> DateTimeTzT:
        dt = _datetime.datetime.strptime(date_string, __format)
        return cls.from_datetime(dt)

    @staticmethod
    def assert_aware_datetime(dt: _datetime.datetime) -> None:
        """
        Check that the given `datetime.datetime` instance is timezone aware. Throws
        an AssertionError otherwise.
        """
        assert dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None

    def __deepcopy__(self: DateTimeTzT, memodict: object) -> DateTimeTzT:
        """
        Deepcopy does not natively work with the __init__ we add to this class
        for extra assertions. Therefore, we override it.
        """
        return self.__class__(
            year=self.year,
            month=self.month,
            day=self.day,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            tzinfo=self.tzinfo,  # type: ignore[arg-type]
        )


class datetime_utc(datetime_tz):
    """
    A `datetime_utc` is a `datetime_tz` but which is guaranteed to be in the UTC+0 timezone.
    """

    assumed_timezone_for_timezone_naive_input = ZoneInfo('UTC')

    @classmethod
    def fromtimestamp(cls, timestamp: float) -> datetime_utc:  # type: ignore[override]  # pylint: disable=arguments-differ
        """
        Parses a timestamp to a timezone aware datetime.
        """
        return cls.from_datetime(_datetime.datetime.fromtimestamp(timestamp, tz=ZoneInfo('UTC')))


class datetime_local(datetime_tz):
    """
    A `datetime_local` is a `datetime_tz` but which is guaranteed to be in the local timezone.
    """

    assumed_timezone_for_timezone_naive_input = tz_local



from zoneinfo import ZoneInfo

from heliclockter import datetime_utc, timedelta


def test_datetime_tz_subtraction() -> None:
    dt_1 = datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC'))
    dt_2 = datetime_utc(2020, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC'))

    assert dt_1 - dt_2 == timedelta(days=366)
    assert dt_2 - dt_1 == timedelta(days=-366)

    assert dt_1 - timedelta(days=366) == dt_2



from copy import deepcopy
from datetime import datetime, timezone, tzinfo
from typing import Optional, Type, Union
from zoneinfo import ZoneInfo

import pytest
from parameterized import parameterized  # type: ignore[import]

from heliclockter import (
    DateTimeTzT,
    DatetimeTzError,
    datetime_local,
    datetime_tz,
    datetime_utc,
    timedelta,
    tz_local,
)

from tests.shared import datetime_cet

DatetimeT = Union[Type[datetime_tz], Type[datetime_cet], Type[datetime_utc]]


@parameterized.expand(
    [
        # Tests for datetime_tz. Parsing datetime_tz from `datetime` should not modify hours.
        (ZoneInfo('CET'), datetime_tz, 10),
        (timezone(timedelta(hours=1)), datetime_tz, 10),
        (ZoneInfo('UTC'), datetime_tz, 10),
        (timezone(timedelta(hours=-4)), datetime_tz, 10),
        (timezone(timedelta(hours=4)), datetime_tz, 10),
        # Tests for datetime_cet. Parsing datetime_cet from `datetime` should modify hours, if
        # the UTC offset of the `datetime` is not equal to that of the 'CET' timezone.
        (ZoneInfo('CET'), datetime_cet, 10),
        (timezone(timedelta(hours=1)), datetime_cet, 10),
        (ZoneInfo('UTC'), datetime_cet, 11),
        (timezone(timedelta(hours=-4)), datetime_cet, 15),
        (timezone(timedelta(hours=4)), datetime_cet, 7),
        # Tests for datetime_utc. Parsing datetime_utc from `datetime` should modify hours, if
        # the timezone of the `datetime` is not UTC.
        (ZoneInfo('CET'), datetime_utc, 9),
        (timezone(timedelta(hours=1)), datetime_utc, 9),
        (ZoneInfo('UTC'), datetime_utc, 10),
        (timezone(timedelta(hours=-4)), datetime_utc, 14),
        (timezone(timedelta(hours=4)), datetime_utc, 6),
        # Tests to ensure a naive datetime can be parsed as datetime_utc, and datetime_cet.
        # We cannot test `datetime_local` as the expected hour will vary on the system testing.
        (None, datetime_utc, 10),
        (None, datetime_cet, 10),
    ]
)
def test_from_datetime(
    input_tz: Optional[tzinfo],
    expected_dt_class: DatetimeT,
    expected_hour: int,
) -> None:
    input_dt = datetime(year=2021, month=1, day=1, hour=10, tzinfo=input_tz)
    dt = expected_dt_class.from_datetime(input_dt)
    assert isinstance(dt, expected_dt_class)
    assert dt.hour == expected_hour


def test_datetime_local_from_datetime() -> None:
    input_dt = datetime(year=2021, month=1, day=1, hour=10)
    dt = datetime_local.from_datetime(input_dt)
    assert isinstance(dt, datetime_local)


def test_datetime_tz_from_naive_datetime() -> None:
    # datetime_tz should not be able to parse from a naive datetime.datetime
    input_dt = datetime(year=2021, month=1, day=1, hour=10)
    with pytest.raises(
        DatetimeTzError, match='Cannot create aware datetime from naive if no tz is assumed'
    ):
        datetime_tz.from_datetime(input_dt)


def test_datetime_tz_now() -> None:
    with pytest.raises(
        DatetimeTzError,
        match=(
            'Must override assumed_timezone_for_timezone_naive_input '
            'or give a timezone when calling now'
        ),
    ):
        datetime_tz.now()

    dt = datetime_tz.now(tz_local)
    assert isinstance(dt, datetime_tz)


def test_datetime_cet_now() -> None:
    dt = datetime_cet.now()
    assert isinstance(dt, datetime_cet)


def test_deepcopy_datetime_tz() -> None:
    dt = datetime_tz(year=2021, month=1, day=1, tzinfo=timezone(timedelta(hours=1)))
    dt_deep_copied = deepcopy(dt)
    assert id(dt) != id(dt_deep_copied)
    assert isinstance(dt_deep_copied, datetime_tz)


@parameterized.expand(
    [
        # UTC naive tests
        (
            '2021-01-10 09:00',
            '%Y-%m-%d %H:%M',
            datetime_utc,
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
        ),
        (
            '2021-01-10 09:00:00',
            '%Y-%m-%d %H:%M:%S',
            datetime_utc,
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
        ),
        (
            '2021-01-10T09:00:00.000000',
            '%Y-%m-%dT%H:%M:%S.%f',
            datetime_utc,
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
        ),
        # UTC aware tests
        (
            '2021-01-10 09:00+00:00',
            '%Y-%m-%d %H:%M%z',
            datetime_utc,
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
        ),
        (
            '2021-01-10 09:00:00+00:00',
            '%Y-%m-%d %H:%M:%S%z',
            datetime_utc,
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
        ),
        (
            '2021-01-10T09:00:00.000000+00:00',
            '%Y-%m-%dT%H:%M:%S.%f%z',
            datetime_utc,
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
        ),
        # CET tests
        (
            '2021-01-10 09:00',
            '%Y-%m-%d %H:%M',
            datetime_cet,
            datetime_cet(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('CET')),
        ),
        (
            '2021-01-10 09:00:00',
            '%Y-%m-%d %H:%M:%S',
            datetime_cet,
            datetime_cet(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('CET')),
        ),
        (
            '2021-01-10T09:00:00.000000',
            '%Y-%m-%dT%H:%M:%S.%f',
            datetime_cet,
            datetime_cet(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('CET')),
        ),
        # CET aware tests
        (
            '2021-01-10 09:00+00:00',
            '%Y-%m-%d %H:%M%z',
            datetime_cet,
            datetime_cet(2021, 1, 10, 10, 00, 00, tzinfo=ZoneInfo('CET')),
        ),
        (
            '2021-01-10 09:00:00+00:00',
            '%Y-%m-%d %H:%M:%S%z',
            datetime_cet,
            datetime_cet(2021, 1, 10, 10, 00, 00, tzinfo=ZoneInfo('CET')),
        ),
        (
            '2021-01-10T09:00:00.000000+00:00',
            '%Y-%m-%dT%H:%M:%S.%f%z',
            datetime_cet,
            datetime_cet(2021, 1, 10, 10, 00, 00, tzinfo=ZoneInfo('CET')),
        ),
    ]
)
def test_strptime(
    dt_string: str, fmt: str, dt_class: Type[DateTimeTzT], expected_dt: DateTimeTzT
) -> None:
    assert dt_class.strptime(dt_string, fmt) == expected_dt


def test_strptime_datetime_tz_naive_dt_string() -> None:
    with pytest.raises(
        DatetimeTzError, match='Cannot create aware datetime from naive if no tz is assumed'
    ):
        datetime_tz.strptime('2021-01-10 09:00', '%Y-%m-%d %H:%M')


@parameterized.expand(
    [
        (1609491600.0, datetime_utc(2021, 1, 1, 9, 0, tzinfo=ZoneInfo('UTC'))),
        (1625126400.0, datetime_utc(2021, 7, 1, 8, 0, tzinfo=ZoneInfo('UTC'))),
        (1656663915.0, datetime_utc(2022, 7, 1, 8, 25, 15, tzinfo=ZoneInfo('UTC'))),
        (1277972715.0, datetime_utc(2010, 7, 1, 8, 25, 15, tzinfo=ZoneInfo('UTC'))),
    ]
)
def test_fromtimestamp(timestamp: float, expected_dt: datetime_utc) -> None:
    parsed_dt = datetime_utc.fromtimestamp(timestamp)
    assert parsed_dt == expected_dt


@parameterized.expand(
    [
        (datetime_utc, 2),
        (datetime_utc, -2),
        (datetime_utc, 0, 2),
        (datetime_cet, 2),
        (datetime_cet, -2),
        (datetime_cet, 0, 2),
        (datetime_local, 2),
        (datetime_local, -2),
        (datetime_local, 0, 2),
        (datetime_tz, 2, 0, ZoneInfo('EST')),
        (datetime_tz, -2, 0, ZoneInfo('EST')),
        (datetime_tz, 0, 2, ZoneInfo('EST')),
    ]
)
def test_future_and_past(
    dt_class: Type[DateTimeTzT], days: int = 0, weeks: int = 0, tz: Optional[ZoneInfo] = None
) -> None:
    delta = timedelta(days=days, weeks=weeks)
    dt_past = dt_class.past(days=days, weeks=weeks, tz=tz)
    dt_now = dt_class.now(tz=tz)
    dt_future = dt_class.future(days=days, weeks=weeks, tz=tz)

    assert (dt_now - dt_past).days == delta.days
    assert (dt_future - dt_now).days == delta.days


def test_future_and_past_no_tz() -> None:
    error_msg = (
        'Must override assumed_timezone_for_timezone_naive_input '
        'or give a timezone when calling now'
    )
    with pytest.raises(DatetimeTzError, match=error_msg):
        datetime_tz.future(days=2)

    with pytest.raises(DatetimeTzError, match=error_msg):
        datetime_tz.past(days=2)



import datetime
from zoneinfo import ZoneInfo

import pytest
from parameterized import parameterized  # type: ignore[import]

from heliclockter import datetime_tz, timedelta, tz_local


@parameterized.expand(
    [
        (datetime.datetime(2021, 1, 1, 10, tzinfo=tz_local),),
        (datetime.datetime(2021, 1, 1, 10, tzinfo=ZoneInfo('CET')),),
        (datetime.datetime(2021, 1, 1, 10, tzinfo=ZoneInfo('UTC')),),
        (
            datetime.datetime(
                2021, 1, 1, 10, tzinfo=datetime.timezone(datetime.timedelta(hours=-8))
            ),
        ),
        (datetime.datetime(2021, 1, 1, 10, tzinfo=datetime.timezone(timedelta(hours=6))),),
    ]
)
def test_assert_aware_datetime(dt: datetime.datetime) -> None:
    datetime_tz.assert_aware_datetime(dt)


def test_assert_aware_datetime_no_tz() -> None:
    with pytest.raises(AssertionError):
        datetime_tz.assert_aware_datetime(datetime.datetime(2021, 1, 1, 10))



from zoneinfo import ZoneInfo

from parameterized import parameterized  # type: ignore[import]

from heliclockter import datetime_tz, datetime_utc


@parameterized.expand(
    [
        (
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
            '%Y-%m-%d',
            '2021-01-10',
        ),
        (
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
            '%Y-%m-%d %H:%M',
            '2021-01-10 09:00',
        ),
        (
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
            '%Y-%m-%d %H:%M:%S',
            '2021-01-10 09:00:00',
        ),
        (
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
            '%Y-%m-%d %H:%M:%SZ',
            '2021-01-10 09:00:00Z',
        ),
        (
            datetime_utc(2021, 1, 10, 9, 00, 00, tzinfo=ZoneInfo('UTC')),
            '%Y-%m-%dT%H:%M:%S.%f',
            '2021-01-10T09:00:00.000000',
        ),
    ]
)
def test_strftime(dt: datetime_tz, fmt: str, expected_str: str) -> None:
    formatted_str = dt.strftime(fmt)
    assert formatted_str == expected_str
